/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	Fatherhood Development
 *
 * NAME:        gfx_sdl.cpp ( 7DRL Library, C++ )
 *
 * COMMENTS:
 */

#include "gfxengine.h"
#include "gfx_sdl.h"

#include <iostream>
using namespace std;

#include <assert.h>

#include "ptrlist.h"
#include "config.h"

#include "SDL.h"

#define KEY_REPEAT_INITIAL 15
#define KEY_REPEAT_AFTER 7

SDL_Surface	*glbVideoSurface = 0;

bool		 screendirty = false;

u8		*glbMultTable;

//
// Key Queue
//
int	*glbKeyQueue = 0;
int	 glbKeyQueueLen = 0;
int	 glbKeyQueueSize = 0;

int	 glbKeyPusher = 0;
int	 glbKeyPusherRaw = 0;
int	 glbKeyPushTime = 0;

//
// Current frame number
//
volatile int	 glbFrameCount = 0;

//
// Cached font information
//
u8 *glbFontMap;
u8 *glbHexMap;
u8 *glbSquareMap;
int glbSquareFullWidth, glbSquareFullHeight;
int glbHexFullWidth, glbHexFullHeight;
int glbFontOffset[256];
int glbFontFullWidth, glbFontFullHeight;
int glbFontWidth, glbFontHeight;
extern GFX_FadeMode	glbFadeMode;
extern u8		glbFadeAmount;

extern PTRLIST<gfx_bmpoverlay *>	glbOverlayList;

// Lookup table
// Text look up tables...
const char *glbTextTable[] = { 
    "abcdefghij",
    "klmnopqrst",
    "uvwxyz^^^^",
    "ABCDEFGHIJ",
    "KLMNOPQRST",
    "UVWXYZ^^^^",
    "0123456789",
    "!@#$%^&*()",
    "=-\\',./+_|",
    "\"<>?[]{}`~",
    ",.<>;:^^^ ",
    0
};

int
gfxsdl_getframecount()
{
    return glbFrameCount;
}

int
gfxsdl_gettilescroll_x()
{
    // We must go EAST!
    int scroll = -gfxsdl_getframecount() / 4;

    return scroll & 31;
}

int
gfxsdl_gettilescroll_y()
{
    return 0;
}

void
gfxsdl_initlut()
{
    glbMultTable = new u8[256*256];

    for (int i = 0; i < 256; i++)
	for (int j = 0; j < 256; j++)
	{
	    glbMultTable[i*256+j] = (i * j + 127) / 255;
	}
}

inline u8
gfxsdl_mult(u8 a, u8 b)
{
    return glbMultTable[ (((unsigned int) a) << 8) + b ];
}

void
gfxsdl_delay(int frames)
{
    int		start;

    start = gfxsdl_getframecount();

    while (gfxsdl_getframecount() - start < frames)
    {
	gfxsdl_awaitEvent();
    }
}

int
gfxsdl_cookkey(int unicode, int sdlkey, int mod)
{
    if (unicode && unicode < 128)
	return unicode;

    switch (sdlkey)
    {
	// Insert several curses here...  This isn't necessary
	// anywhere except WIndows XP it seems...
	case SDLK_KP0:
	    return '0';
	case SDLK_KP1:
	    return '1';
	case SDLK_KP2:
	    return '2';
	case SDLK_KP3:
	    return '3';
	case SDLK_KP4:
	    return '4';
	case SDLK_KP5:
	    return '5';
	case SDLK_KP6:
	    return '6';
	case SDLK_KP7:
	    return '7';
	case SDLK_KP8:
	    return '8';
	case SDLK_KP9:
	    return '9';
	case SDLK_UP:
	    return GFX_KEYUP;
	case SDLK_DOWN:
	    return GFX_KEYDOWN;
	case SDLK_LEFT:
	    return GFX_KEYLEFT;
	case SDLK_RIGHT:
	    return GFX_KEYRIGHT;

	case SDLK_F1:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF1;
	    return GFX_KEYF1;
	case SDLK_F2:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF2;
	    return GFX_KEYF2;
	case SDLK_F3:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF3;
	    return GFX_KEYF3;
	case SDLK_F4:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF4;
	    return GFX_KEYF4;
	case SDLK_F5:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF5;
	    return GFX_KEYF5;
	case SDLK_F6:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF6;
	    return GFX_KEYF6;
	case SDLK_F7:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF7;
	    return GFX_KEYF7;
	case SDLK_F8:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF8;
	    return GFX_KEYF8;
	case SDLK_F9:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF9;
	    return GFX_KEYF9;
	case SDLK_F10:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF10;
	    return GFX_KEYF10;
	case SDLK_F11:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF11;
	    return GFX_KEYF11;
	case SDLK_F12:
	    if (mod & KMOD_SHIFT)
		return GFX_KEYSF12;
	    return GFX_KEYF12;
    }
    return 0;
}

//
// Finds the offset for the given character.
//
int
gfxsdl_findoffset(u8 c)
{
    if (!c)
	return -1;

    int		x, y;

    for (y = 0; y < 11; y++)
    {
	for (x = 0; x < 10; x++)
	{
	    if (c == glbTextTable[y][x])
	    {
		return (x * glbFontWidth + y * glbFontHeight * glbFontFullWidth) * 3;
	    }
	}
    }
    // Not found.
    return -1;
}

void
gfxsdl_initfont()
{
    int	    i;

    glbFontMap = gfx_loadbitmap("../gfx/alphabet_shadow.bmp", glbFontFullWidth, glbFontFullHeight);

    glbSquareMap = gfx_loadbitmap("../gfx/dungeon32.bmp", glbSquareFullWidth, glbSquareFullHeight);
    glbHexMap = gfx_loadbitmap("../gfx/hex32.bmp", glbHexFullWidth, glbHexFullHeight);

    // Determine size of characters.
    glbFontWidth = glbFontFullWidth / 10;
    glbFontHeight = glbFontFullHeight / 11;

    for (i = 0; i < 256; i++)
    {
	glbFontOffset[i] = gfxsdl_findoffset(i);
	// Rather than error out on unknown chars, output ?
	if (glbFontOffset[i] < 0)
	    glbFontOffset[i] = gfxsdl_findoffset('?');
    }
}

Uint32
glb_VBLTimerCallback(Uint32 interval, void *fp)
{
    glbFrameCount++;

    // Push a fake event so we wake up.
    SDL_Event		event;
    event.type = SDL_USEREVENT;
    SDL_PushEvent(&event);

    return 16;
}

void
gfxsdl_init()
{
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
    {
	cerr << "Video init failed." << endl;
	SDL_Quit();
	exit(-1);
    }

    gfxsdl_initfont();

    // Clear our text array
    gfx_clearbox(0, 0, SCR_WIDTH, SCR_HEIGHT, ' ', ATTR_NONE);

    int		flags = 0;
    if (glbConfig->screenFull())
	flags = SDL_FULLSCREEN;

    glbVideoSurface = SDL_SetVideoMode(SCR_WIDTH * glbFontWidth, SCR_HEIGHT * glbFontHeight, 24, 
	    flags);
    if (!glbVideoSurface)
    {
	cerr << "Failed to acquire video surface." << endl;
	cerr << "Attempted resolution was " << (SCR_WIDTH * glbFontWidth) << " by " << (SCR_HEIGHT * glbFontHeight) << endl;
	cerr << "SDL Error is: " << SDL_GetError() << endl;
	SDL_Quit();
	exit(-1);
    }

    SDL_WM_SetCaption("Explore! Malachite Dreams", 0);

    SDL_EnableUNICODE(1);

    SDL_AddTimer(16, glb_VBLTimerCallback, 0);

    gfxsdl_initlut();
}

void
gfxsdl_shutdown()
{
    delete [] glbFontMap;
    glbFontMap = 0;
    SDL_Quit();
}

void
gfxsdl_blitchar(u8 *dst, int pitch, u8 c, int attr)
{
    // Colours.
    int		 offset;
    u8		*src, r, g, b;
    int		 x, y;

    src = glbFontMap;
    offset = glbFontOffset[c];
    if (offset < 0)
    {
	// Invalid character.  Ignore.
	assert(!"Unknown character.");
	return;
    }
    src += offset;

    if (glb_attrdefs[attr].transparent)
    {
	// Special case transparent spaces as there are a lot of them
	if (c == ' ')
	    return;
	if (glb_attrdefs[attr].remap)
	{
	    for (y = 0; y < glbFontHeight; y++)
	    {
		for (x = 0; x < glbFontWidth; x++)
		{
		    r = *src++;
		    g = *src++;
		    b = *src++;

		    // All black, transparent.
		    if ((r | g | b))
		    {
			if ((r & g & b) == 0xff)
			{
			    *dst++ = glb_attrdefs[attr].fg_b;
			    *dst++ = glb_attrdefs[attr].fg_g;
			    *dst++ = glb_attrdefs[attr].fg_r;
			}
			else
			{
			    // Background
			    *dst++ = glb_attrdefs[attr].bg_b;
			    *dst++ = glb_attrdefs[attr].bg_g;
			    *dst++ = glb_attrdefs[attr].bg_r;
			}
		    }
		    else
		    {
			dst += 3;
		    }
		}
		src += (glbFontFullWidth - glbFontWidth) * 3;
		dst += pitch - 3 * glbFontWidth;
	    }
	}
	else
	{
	    for (y = 0; y < glbFontHeight; y++)
	    {
		for (x = 0; x < glbFontWidth; x++)
		{
		    r = *src++;
		    g = *src++;
		    b = *src++;

		    // All black, transparent.
		    if ((r | g | b))
		    {
			*dst++ = r;
			*dst++ = g;
			*dst++ = b;
		    }
		    else
		    {
			dst += 3;
		    }
		}
		src += (glbFontFullWidth - glbFontWidth) * 3;
		dst += pitch - 3 * glbFontWidth;
	    }
	}
    }
    else if (glb_attrdefs[attr].remap)
    {
	for (y = 0; y < glbFontHeight; y++)
	{
	    for (x = 0; x < glbFontWidth; x++)
	    {
		r = *src++;
		g = *src++;
		b = *src++;

		// All white, use foreground.  Otherwise
		// background.
		if ((r & g & b) == 255)
		{
		    *dst++ = glb_attrdefs[attr].fg_b;
		    *dst++ = glb_attrdefs[attr].fg_g;
		    *dst++ = glb_attrdefs[attr].fg_r;
		}
		else
		{
		    *dst++ = glb_attrdefs[attr].bg_b;
		    *dst++ = glb_attrdefs[attr].bg_g;
		    *dst++ = glb_attrdefs[attr].bg_r;
		}
	    }
	    src += (glbFontFullWidth - glbFontWidth) * 3;
	    dst += pitch - 3 * glbFontWidth;
	}
    }
    else
    {
	for (y = 0; y < glbFontHeight; y++)
	{
	    for (x = 0; x < glbFontWidth; x++)
	    {
		*dst++ = *src++;
		*dst++ = *src++;
		*dst++ = *src++;
	    }
	    src += (glbFontFullWidth - glbFontWidth) * 3;
	    dst += pitch - 3 * glbFontWidth;
	}
    }
}

const int glb_hexoffset[32] =
{
    22,
    21,
    21,
    20,
    19,
    19,
    18,
    17,
    17,
    16,
    15,
    15,
    14,
    13,
    13,
    12,

    12,
    13,
    13,
    14,
    15,
    15,
    16,
    17,
    17,
    18,
    19,
    19,
    20,
    21,
    21,
    22
};

void
gfxsdl_blittile(u8 *dst, int hexx, int hexy, int stride, TILE_NAMES tile)
{
    // Compute screen coords
    int			x, y;

    x = hexx * 32;
    y = hexy * 32;

    {
	// Blit as a square.
	u8		*src = glbSquareMap;

	src += (glb_tiledefs[tile].x * 32 + glb_tiledefs[tile].y * 32 * glbSquareFullWidth) * 3;
	dst += x * 3 + y * stride;
	if (glb_tiledefs[tile].shadow)
	{
	    for (int iy = 0; iy < 32; iy++)
	    {
		for (int ix = 0; ix < 32; ix++)
		{
		    *dst++ = gfxsdl_mult(*src, *dst);
		    src++;
		    *dst++ = gfxsdl_mult(*src, *dst);
		    src++;
		    *dst++ = gfxsdl_mult(*src, *dst);
		    src++;
		}
		src += glbSquareFullWidth * 3 - 32*3;
		dst += stride - 32 * 3;
	    }
	}
	else if (glb_tiledefs[tile].transparent)
	{
	    for (int iy = 0; iy < 32; iy++)
	    {
		for (int ix = 0; ix < 32; ix++)
		{
		    if (!*src && !src[1] && !src[2])
		    {
			src += 3;
			dst += 3;
		    }
		    else
		    {
			*dst++ = *src++;
			*dst++ = *src++;
			*dst++ = *src++;
		    }
		}
		src += glbSquareFullWidth * 3 - 32*3;
		dst += stride - 32 * 3;
	    }
	}
	else if (glb_tiledefs[tile].scrolling)
	{
	    // Send it EAST!
	    int		scrolloffx = gfxsdl_gettilescroll_x();
	    int		scrolloffy = gfxsdl_gettilescroll_y();
	    scrolloffy = 0;

	    scrolloffy &= 31;
	    scrolloffx &= 31;
	    const u8	*srcline;
	    for (int iy = 0; iy < 32; iy++)
	    {
		srcline = src + glbSquareFullWidth * 3 * ( (iy + scrolloffy) & 31 );
		for (int ix = 0; ix < 32; ix++)
		{
		    const u8 *srcpixel = &srcline[ ((ix + scrolloffx) & 31) * 3];
		    *dst++ = *srcpixel++;
		    *dst++ = *srcpixel++;
		    *dst++ = *srcpixel++;
		}
		dst += stride - 32 * 3;
	    }
	}
	else
	{
	    for (int iy = 0; iy < 32; iy++)
	    {
		for (int ix = 0; ix < 32; ix++)
		{
		    *dst++ = *src++;
		    *dst++ = *src++;
		    *dst++ = *src++;
		}
		src += glbSquareFullWidth * 3 - 32*3;
		dst += stride - 32 * 3;
	    }
	}
    }
}

void
gfxsdl_extracttile(u8 *dst,
		   const u8 *src, int sx, int sy, int srcstride)
{
    src += sx * 32 * 3 + sy * 32 * srcstride;
    for (int y = 0; y < 32; y++)
    {
	for (int x = 0; x < 32; x++)
	{
	    *dst++ = *src++;
	    *dst++ = *src++;
	    *dst++ = *src++;
	}
	src += srcstride - 32*3;
    }
}

void
gfxsdl_blitmaskedtile(u8 *dst, int hexx, int hexy, int stride, u8 *origtile, TILE_NAMES tile, TILE_NAMES masktile)
{
    // Compute screen coords
    int			x, y;

    x = hexx * 32;
    y = hexy * 32;

    {
	// Blit as a square.
	u8		*src = glbSquareMap;
	u8		*mask = glbSquareMap;

	src += (glb_tiledefs[tile].x * 32 + glb_tiledefs[tile].y * 32 * glbSquareFullWidth) * 3;
	mask += (glb_tiledefs[masktile].x * 32 + glb_tiledefs[masktile].y * 32 * glbSquareFullWidth) * 3;
	dst += x * 3 + y * stride;


	if (glb_tiledefs[tile].shadow)
	{
	    for (int iy = 0; iy < 32; iy++)
	    {
		for (int ix = 0; ix < 32; ix++)
		{
		    if (!*mask && !mask[1] && !mask[2])
		    {
			src += 3;
			dst += 3;
			mask += 3;
			origtile += 3;
		    }
		    else
		    {
			*dst++ = gfxsdl_mult(*src, *origtile);
			src++;
			origtile++;
			*dst++ = gfxsdl_mult(*src, *origtile);
			src++;
			origtile++;
			*dst++ = gfxsdl_mult(*src, *origtile);
			src++;
			origtile++;
			mask += 3;
		    }
		}
		src += glbSquareFullWidth * 3 - 32*3;
		mask += glbSquareFullWidth * 3 - 32*3;
		dst += stride - 32 * 3;
		// Assume its stride matches.
		origtile += 0;
	    }
	}
	else if (glb_tiledefs[tile].scrolling)
	{
	    // Send it EAST!
	    int		scrolloffx = gfxsdl_gettilescroll_x();
	    int		scrolloffy = gfxsdl_gettilescroll_y();
	    scrolloffy = 0;

	    scrolloffy &= 31;
	    scrolloffx &= 31;
	    const u8	*srcline;
	    for (int iy = 0; iy < 32; iy++)
	    {
		srcline = src + glbSquareFullWidth * 3 * ( (iy + scrolloffy) & 31 );
		for (int ix = 0; ix < 32; ix++)
		{
		    const u8 *srcpixel = &srcline[ ((ix + scrolloffx) & 31) * 3];
		    if (!*mask && !mask[1] && !mask[2])
		    {
			dst += 3;
			mask += 3;
		    }
		    else
		    {
			*dst++ = *srcpixel++;
			*dst++ = *srcpixel++;
			*dst++ = *srcpixel++;
			mask += 3;
		    }
		}
		dst += stride - 32 * 3;
		mask += glbSquareFullWidth * 3 - 32*3;
	    }
	}
	else
	{
	    // Normal!
	    for (int iy = 0; iy < 32; iy++)
	    {
		for (int ix = 0; ix < 32; ix++)
		{
		    if (!*mask && !mask[1] && !mask[2])
		    {
			src += 3;
			dst += 3;
			mask += 3;
		    }
		    else
		    {
			*dst++ = *src++;
			*dst++ = *src++;
			*dst++ = *src++;
			mask += 3;
		    }
		}
		src += glbSquareFullWidth * 3 - 32*3;
		mask += glbSquareFullWidth * 3 - 32*3;
		dst += stride - 32 * 3;
	    }
	}
    }
}
TILE_NAMES
gfxsdl_bleedmaskfromdxdy(int x, int y, int dx, int dy)
{
    if (dx < 0 && dy < 0)
	return TILE_MASK_BLEED_TOPLEFT;
    if (dx < 0 && dy > 0)
	return TILE_MASK_BLEED_BOTLEFT;
    if (dx > 0 && dy < 0)
	return TILE_MASK_BLEED_TOPRIGHT;
    if (dx > 0 && dy > 0)
	return TILE_MASK_BLEED_BOTRIGHT;

    if (dx < 0)
	return TILE_MASK_BLEED_LEFT;
    if (dx > 0)
	return TILE_MASK_BLEED_RIGHT;
    if (dy < 0)
	return TILE_MASK_BLEED_TOP;
    if (dy > 0)
	return TILE_MASK_BLEED_BOT;

    return TILE_NONE;
}

void
gfxsdl_computeanddrawbleed(u8 *dst, int stride, int layer, int x, int y)
{
    TILE_NAMES tile = gfx_gettile(layer, x, y);
    TILE_NAMES srctile, srcbleed, mask;

    if (!glb_tiledefs[tile].allowbleed)
	return;

    int		priority = glb_tiledefs[tile].bleed_priority;
    u8		origtile[32*32 * 3];

    // We must always do blends against the un-bled tile to avoid
    // double comping shadows!
    gfxsdl_extracttile(origtile, dst, x, y, stride);

    int		dx, dy;
    FORALL_8DIR(dx, dy)
    {
	if (x + dx < 0) continue;
	if (x + dx >= SCR_WIDTH/4) continue;
	if (y + dy < 0) continue;
	if (y + dy >= SCR_HEIGHT/4) continue;

	if (!dx || !dy)
	{
	    srctile = gfx_gettile(layer, x+dx, y+dy);
	    srcbleed = (TILE_NAMES) glb_tiledefs[srctile].bleed;
	    if (srcbleed == TILE_NONE)
		continue;
	    
	    // Verify its priority is strictly greater than ours!
	    if (glb_tiledefs[srctile].bleed_priority > priority)
	    {
		// Render on top!
		mask = gfxsdl_bleedmaskfromdxdy(x, y, dx, dy);
		if (mask != TILE_NONE)
		{
		    gfxsdl_blitmaskedtile(dst, x, y, stride, origtile, srcbleed, mask);
		}
	    }
	}
	else
	{
	    // We have a corner test.
	    TILE_NAMES		xsrc = gfx_gettile(layer, x+dx, y);
	    TILE_NAMES		ysrc = gfx_gettile(layer, x, y + dy);

	    if (glb_tiledefs[xsrc].bleed != glb_tiledefs[ysrc].bleed)
		continue;

	    srcbleed = (TILE_NAMES) glb_tiledefs[xsrc].bleed;
	    if (srcbleed == TILE_NONE)
		continue;

	    if (glb_tiledefs[xsrc].bleed_priority > priority &&
		glb_tiledefs[ysrc].bleed_priority > priority)
	    {
		// Render on top!
		mask = gfxsdl_bleedmaskfromdxdy(x, y, dx, dy);
		if (mask != TILE_NONE)
		{
		    gfxsdl_blitmaskedtile(dst, x, y, stride, origtile, srcbleed, mask);
		}
	    }
	}
    }
}

void
gfxsdl_drawoverlays(int layer)
{
    for (int overlay = 0; overlay < glbOverlayList.entries(); overlay++)
    {
	gfx_bmpoverlay *o = glbOverlayList(overlay);
	if (!o)
	    continue;

	if (o->layer != layer)
	    continue;

	for (int y = o->y; y < o->y+o->h*o->scale; y++)
	{
	    if (y < 0)
		continue;
	    if (y >= SCR_HEIGHT*glbFontHeight)
		continue;

	    u8	*dst = (u8 *) glbVideoSurface->pixels;
	    dst += glbVideoSurface->pitch * y;

	    const u8 *rgb = &o->rgbdata[ ((y - o->y)/o->scale) * o->w * 3];
	    if (o->x < 0)
	    {
		rgb += 3 * ((-o->x) / o->scale);
	    }
	    else
	    {
		dst += o->x * 3;
	    }

	    int		ex = MIN(o->x+o->w*o->scale, SCR_WIDTH*glbFontWidth);
	    u8		r, g, b;
	    int		scaleoff = 0;
	    if (o->transparent)
	    {
		for (int x = MAX(o->x, 0); x < ex; x++)
		{
		    r = *rgb++;
		    g = *rgb++;
		    b = *rgb++;
		    scaleoff++;
		    if (scaleoff < o->scale)
		    {
			rgb -= 3;
		    }
		    else
		    {
			scaleoff = 0;
		    }
		    if ((r | g | b))
		    {
			*dst++ = r;
			*dst++ = g;
			*dst++ = b;
		    }
		    else
		    {
			dst += 3;
		    }
		}
	    }
	    else
	    {
		for (int x = MAX(o->x, 0); x < ex; x++)
		{
		    r = *rgb++;
		    g = *rgb++;
		    b = *rgb++;

		    scaleoff++;
		    if (scaleoff < o->scale)
		    {
			rgb -= 3;
		    }
		    else
		    {
			scaleoff = 0;
		    }
		    *dst++ = r;
		    *dst++ = g;
		    *dst++ = b;
		}
	    }
	}
    }
}

void
gfxsdl_rebuildscreen()
{
    int		x, y;
    u8		c;
    ATTR_NAMES	attr;

    SDL_LockSurface(glbVideoSurface);

    u8		*dst;
    dst = (u8 *) glbVideoSurface->pixels;
    if (!dst)
    {
	cerr << "Lock failure: " << SDL_GetError() << endl;
    }

    // Wipe to black.
    memset(dst, 0, SCR_HEIGHT * glbFontHeight * glbVideoSurface->pitch);

    // Draw in background bitmaps
    gfxsdl_drawoverlays(GFX_OVERLAY_BACK);

    // Draw mid layer tiles.
    for (y = 0; y < SCR_HEIGHT/4; y++)
    {
	for (x = 0; x < SCR_WIDTH/4; x++)
	{
	    // We put the layer on the inside as this is then
	    // most cache friendly with our output.
	    for (int layer = 0; layer < 4; layer++)
	    {
		TILE_NAMES tile = gfx_gettile(layer, x, y);

		if (tile != TILE_NONE)
		{
		    gfxsdl_blittile(dst, x, y, glbVideoSurface->pitch, tile);

		    gfxsdl_computeanddrawbleed(dst, glbVideoSurface->pitch, layer, x, y);
		}
	    }
	}
    }

    // Draw mid level bitmaps
    gfxsdl_drawoverlays(GFX_OVERLAY_MID);

    // Draw top layer tiles.
    for (y = 0; y < SCR_HEIGHT/4; y++)
    {
	for (x = 0; x < SCR_WIDTH/4; x++)
	{
	    // We put the layer on the inside as this is then
	    // most cache friendly with our output.
	    for (int layer = 4; layer < MAX_LAYER; layer++)
	    {
		TILE_NAMES tile = gfx_gettile(layer, x, y);

		if (tile != TILE_NONE)
		{
		    gfxsdl_blittile(dst, x, y, glbVideoSurface->pitch, tile);
		    gfxsdl_computeanddrawbleed(dst, glbVideoSurface->pitch, layer, x, y);
		}
	    }
	}
    }

    if (glbFadeMode)
    {
	for (y = 0; y < SCR_HEIGHT*glbFontHeight; y++)
	{
	    u8	*dst = (u8 *) glbVideoSurface->pixels;
	    dst += glbVideoSurface->pitch * y;
	    for (x = 0; x < SCR_WIDTH*glbFontHeight; x++)
	    {
		*dst = gfxsdl_mult(glbFadeAmount, *dst);
		dst++;
		*dst = gfxsdl_mult(glbFadeAmount, *dst);
		dst++;
		*dst = gfxsdl_mult(glbFadeAmount, *dst);
		dst++;
	    }
	}
    }
#if 1
    for (y = 0; y < SCR_HEIGHT; y++)
    {
	for (x = 0; x < SCR_WIDTH; x++)
	{
	    gfx_getchar(x, y, c, attr);

	    if (c)
		gfxsdl_blitchar(&dst[x*glbFontWidth*3+y*glbFontHeight*glbVideoSurface->pitch], glbVideoSurface->pitch, c, attr);
	}
    }
#endif

    // Draw top level bitmaps
    gfxsdl_drawoverlays(GFX_OVERLAY_TOP);

    SDL_UnlockSurface(glbVideoSurface);

    SDL_UpdateRect(glbVideoSurface, 0, 0, 0, 0);
}

//
// Event Polling
//
void
gfxsdl_pollEvents()
{
    SDL_Event		event;

    while (SDL_PollEvent(&event))
    {
	switch (event.type)
	{
	    case SDL_KEYDOWN:
		// For now, ignore arrow keys..
		int key;

		key = gfxsdl_cookkey(event.key.keysym.unicode, event.key.keysym.sym, event.key.keysym.mod);
		// If the key doesn't cook to anything interesting, don't
		// add it to our list.
		if (key)
		    gfxsdl_putKey(key);
		// Set the pusher.
		glbKeyPusher = key;
		glbKeyPusherRaw = event.key.keysym.sym;
		glbKeyPushTime = gfxsdl_getframecount() + KEY_REPEAT_INITIAL;
		break;

	    case SDL_KEYUP:
		// Stop the pusher.
		// Only do this if the raw pusher symbol matches this key
		// up, preventing the keyboard repeat from being stopped
		// by the previous key going high.
		// We cannot use glbKeyPusher because unicode isn't cooked
		// on a key up.
		if (glbKeyPusherRaw == event.key.keysym.sym)
		{
		    glbKeyPusher = 0;
		}
		break;

	    case SDL_QUIT:
		SDL_Quit();
		exit(0);
		break;
	}
    }

    // Note that we handle key repeat here, not in the actual timer loop!
    // This way if our processing freezes, the key repeating also
    // freezes.
    {
	int		frame;

	frame = gfxsdl_getframecount();
	if (glbKeyPusher)
	{
	    if (frame >= glbKeyPushTime)
	    {
		gfxsdl_putKey(glbKeyPusher);
		glbKeyPushTime = frame + KEY_REPEAT_AFTER;
	    }
	}
    }
}

//
// Keyboard functions
//
bool
gfxsdl_isKeyWaiting()
{
    gfxsdl_pollEvents();

    if (glbKeyQueueLen > 0)
	return true;
    return false;
}

void
gfxsdl_clearKeyBuffer()
{
    gfxsdl_pollEvents();
    glbKeyQueueLen = 0;
}

void
gfxsdl_putKey(int key)
{
    if (glbKeyQueueLen >= glbKeyQueueSize)
    {
	// Grow the queue.
	int		*newqueue;

	newqueue = new int [(glbKeyQueueSize + 4)*2];

	if (glbKeyQueueLen)
	{
	    memcpy(newqueue, glbKeyQueue, sizeof(int) * glbKeyQueueLen);
	}
	delete [] glbKeyQueue;
	glbKeyQueue = newqueue;
	glbKeyQueueSize = (glbKeyQueueSize + 4) * 2;
    }

    glbKeyQueue[glbKeyQueueLen++] = key;
}

void
gfxsdl_breakKeyRepeat()
{
    glbKeyPusher = 0;
}

int
gfxsdl_getKey()
{
    if (!gfxsdl_isKeyWaiting())
	return 0;

    int		key;

    key = glbKeyQueue[0];
    memmove(glbKeyQueue, &glbKeyQueue[1], sizeof(int) * (glbKeyQueueLen-1));

    glbKeyQueueLen--;
    return key;
}

void
gfxsdl_awaitEvent()
{
    SDL_WaitEvent(0);
}
