#include "glbdef.h"
#include "gfxengine.h"
#include <stdio.h>
#include <iostream>
using namespace std;

#include "map.h"
#include "atlas.h"
#include "build.h"
#include "mob.h"
#include "panel.h"
#include "text.h"
#include "runes.h"
#include "config.h"
#include "rooms/allrooms.h"

MOB *glbAvatar = 0;
PANEL *glbMsg = 0;
PANEL *glbTalkBox = 0;
PANEL *glbBook = 0;

void
redrawWorld()
{
    gfx_clearbox(0, 0, SCR_WIDTH, SCR_HEIGHT, ' ', ATTR_NORMAL);
    if (glbAvatar)
	MAP::render(glbAvatar->pos(), glbAvatar->getFacing(), glbAvatar);
    glbMsg->redraw();
    glbTalkBox->redraw();
    glbBook->redraw();
    gfx_update();
}

void
fadeToBlack()
{
    redrawWorld();
    gfx_fadeto(GFX_FADE_BLACK, false);
    gfx_clearKeyBuffer();
}

void
fadeFromBlack()
{
    redrawWorld();
    gfx_fadeto(GFX_FADE_BLACK, true);
    gfx_setfade(GFX_FADE_NONE, 0);
    gfx_clearKeyBuffer();
}

void
readBook(const char *text)
{
    glbBook->clear();
    glbBook->appendText(text);
    int		height;

    height = (glbBook->getCurLine() + 3) / 4;
    for (int x = 0; x < 10; x++)
    {
	gfx_settile(4, x + 10, 2, TILE_PAPER_TOP);
	for (int y = 3; y < 3 + height; y++)
	{
	    gfx_settile(4, x + 10, y, TILE_PAPER);
	}
	gfx_settile(4, x + 10, 3 + height, TILE_PAPER_BOTTOM);
    }
    redrawWorld();
    gfx_getKey(true);

    for (int x = 0; x < 10; x++)
    {
	for (int y = 2; y < 4 + height; y++)
	{
	    gfx_settile(4, x + 10, y, TILE_NONE);
	}
    }
    glbBook->clear();
}
void
readBook(BUF buf) { readBook(buf.buffer()); }

void
gba_main()
{
    glbConfig = new CONFIG();
    glbConfig->load("../explore.cfg");

    gfx_init();
    text_init();

    glbMsg = new PANEL(40, 20);
    glbMsg->move(3, 3);
    glbMsg->setTransparent(true);
    glbTalkBox = new PANEL(100, 1);
    glbTalkBox->move(3, 50);
    glbTalkBox->setTransparent(true);
    glbBook = new PANEL(39, 50);
    glbBook->move(41, 12);
    glbBook->setTransparent(true);
    glbBook->setIndent(1);

    gfx_setfade(GFX_FADE_BLACK, 0);

    glbMsg->move(20, 20);

    // Welcome
    glbMsg->appendText(text_lookup("game", "welcome"));
    redrawWorld();
    gfx_getKey(true);
    glbMsg->erase();
    glbMsg->clear();

    // Validation of our data sets!
    {
	POI_NAMES	poi;
	FOREACH_POI(poi)
	{
	    if (poi == POI_NONE)
		continue;
	    if (!text_hasentry("poi", glb_poidefs[poi].name))
	    {
		printf("POI %s missing entry.\n",
			glb_poidefs[poi].name);
		assert(!"Missing POI text");
	    }
	}

	SPELL_NAMES	spell;
	FOREACH_SPELL(spell)
	{
	    if (spell == SPELL_NONE)
		continue;
	    if (!text_hasentry("spell", glb_spelldefs[spell].name))
	    {
		printf("Spell %s missing entry.\n",
			glb_spelldefs[spell].name);
		assert(!"Missing Spell text");
	    }
	}
    }

    // Fetch the name!
    while (1)
    {
	char		buf[100];
	glbMsg->appendText("Supplicant!\n\nSpeak your true name!\n");
	glbMsg->getString("> ", buf, 20);

	BUF		name;
	name.reference(buf);
	name.stripWS();

	if (!name.isstring())
	{
	    // Sorry, be less lazy!
	    glbMsg->appendText("\n");
	    glbMsg->appendText(text_lookup("game", "emptyname"));
	    redrawWorld();
	    gfx_getKey(true);
	    glbMsg->erase();
	    glbMsg->clear();

	    continue;
	}

	if (name.strchr('$'))
	{
	    // Safety first!
	    glbMsg->appendText("\n");
	    glbMsg->appendText(text_lookup("game", "dollarname"));
	    redrawWorld();
	    gfx_getKey(true);
	    glbMsg->erase();
	    glbMsg->clear();

	    continue;
	}
	
	// Valid name!
	glbMsg->erase();
	glbMsg->clear();

	text_registervariable("name", name);

	rand_setseed(name.hash());

	break;
    }


    glbMsg->appendText("A world of dreams slowly forms...\nHave patience.");
    redrawWorld();

    MAP		*map;
    ATLAS	*world = glbAtlas();

    RUNES::getRunes()->init();

    // Register our variables.
    {
	SPELL_NAMES spell;
	FOREACH_SPELL(spell)
	{
	    text_registervariable(glb_spelldefs[spell].words, RUNES::getRunes()->spellRune(spell));
	}

	POI_NAMES poi;
	FOREACH_POI(poi)
	{
	    text_registervariable(glb_poidefs[poi].name, RUNES::getRunes()->diskRune(poi));
	}

	text_registervariable("answer", RUNES::getRunes()->riddleRune());
	// Pick a random meaning.
	BUF		meaning;
	const char *possiblemeanings[3] = { "meaning::wall", "meaning::skull", "meaning::archway" };
	meaning = text_lookup(possiblemeanings[rand_choice(3)]);
	text_registervariable("meaning", meaning);
    }

    {
	int		 borderw, borderh;
        u8		*border_rgb = gfx_loadbitmap("../gfx/border.bmp", borderw, borderh);

	if (border_rgb)
	{
	    gfx_overlay_create(border_rgb, borderw, borderh, 0, 0, 1, GFX_OVERLAY_MID, true);
	    delete [] border_rgb;
	}
    }

    {
	int		 blazew, blazeh;
        u8		*blaze_rgb = gfx_loadbitmap("../gfx/blaze.bmp", blazew, blazeh);

	if (blaze_rgb)
	{
	    gfx_overlay_create(blaze_rgb, blazew, blazeh, 480, 0, 1, GFX_OVERLAY_BACK, false);
	    delete [] blaze_rgb;
	}
    }

    for (int i = 0; i < NUM_ALLROOMDEFS; i++)
    {
	MAP	*room = new MAP(*glb_allroomdefs[i]);
	world->registerMap(room);
    }

#if 0

    glbAvatar = new MOB(MOB_AVATAR);
    int		 iter = 0;
    while (1)
    {
	//ROOM_DEF	*def = BUILD::build(BUILD_WORLD, 0, "maze", "maze");
	//ROOM_DEF	*def = BUILD::build(BUILD_POI, rand_choice(2) ? POI_PORT : POI_TELE, "maze", "maze");
	POI_NAMES	randpoi = POI_NONE;
	while (randpoi == POI_NONE)
	{
	    randpoi = (POI_NAMES) rand_choice(NUM_POIS);
	    if (randpoi == POI_PORT)
		randpoi = POI_NONE;
	}
	ROOM_DEF	*def = BUILD::build(BUILD_POI, randpoi, "maze", "maze");
	
	MAP		*room = new MAP(*def);
	printf("Iteration %d\n", iter++);

#if 0
	world->registerMap(room);

	POS p = room->findRandomPassable();
	glbAvatar->moveTo(p);

	glbAvatar->actionCast(SPELL_BIRDSVIEW);
#else
	delete room;
	delete (ROOM_DEF_NC *)def;
#endif
    }
#endif

    {
	ROOM_DEF	*def = BUILD::build(BUILD_WORLD, 0, "world", "world");
	MAP		*room = new MAP(*def);
	world->registerMap(room);

	POI_NAMES	poi;
	FOREACH_POI(poi)
	{
	    if (poi == POI_NONE)
		continue;

	    BUF		mapname;
	    mapname.sprintf("%s_gen", glb_poidefs[poi].name);

	    def = BUILD::build(BUILD_POI, poi, mapname.buffer(), 0);
	    room = new MAP(*def);
	    world->registerMap(room);
	}
    }


    // Done world building!
    glbMsg->clear();
    redrawWorld();
    gfx_clearKeyBuffer();


    glbAvatar = new MOB(MOB_AVATAR);

#if 0
    map = world->findMap("world");
    POS p = map->findRandomItem(ITEM_HUT);
#else
    map = world->findMap("home");
    POS p = map->findRandomItem(ITEM_BED);
#endif
    glbAvatar->moveTo(p);
     
    gfx_clearbox(0, 0, SCR_WIDTH, SCR_HEIGHT, '\0', ATTR_NONE);
    bool		done = false;
    bool		didmove;

    BUF			fkeys[12];

    fadeFromBlack();

    while (!done)
    {
	world->buildFOV(glbAvatar->pos(), 10);
	redrawWorld();

	didmove = false;

	{
	    int		key;

	    key = gfx_getKey(false);

	    if (key)
	    {
		world->clearMessages();
		glbTalkBox->erase();
		glbTalkBox->clear();

		switch (key)
		{
		    case 'k':
		    case '8':
		    case GFX_KEYUP:
			didmove = glbAvatar->actionBump(DIR::north() - glbAvatar->getFacing());
			break;
		    case GFX_KEYDOWN:
		    case '2':
		    case 'j':
			didmove = glbAvatar->actionBump(DIR::south() - glbAvatar->getFacing());
			break;
		    case GFX_KEYLEFT:
		    case 'h':
		    case '4':
			didmove = glbAvatar->actionBump(DIR::west() - glbAvatar->getFacing());
			break;
		    case GFX_KEYRIGHT:
		    case 'l':
		    case '6':
			didmove = glbAvatar->actionBump(DIR::east() - glbAvatar->getFacing());
			break;


		    case GFX_KEYF1:
		    case GFX_KEYF2:
		    case GFX_KEYF3:
		    case GFX_KEYF4:
		    case GFX_KEYF5:
		    case GFX_KEYF6:
		    case GFX_KEYF7:
		    case GFX_KEYF8:
		    case GFX_KEYF9:
		    case GFX_KEYF10:
		    case GFX_KEYF11:
		    case GFX_KEYF12:
		    {
			int		whichfunc = key - GFX_KEYF1;

			if (fkeys[whichfunc].isstring())
			{
			    didmove = glbAvatar->actionYell(fkeys[whichfunc].buffer());
			}
			else
			{
			    BUF		bind;
			    bind.sprintf("F%d is unbound: Use Shift-F%d to bind.", whichfunc+1, whichfunc+1);
			    glbTalkBox->appendText(bind);
			    didmove = false;
			}
			break;
		    }

		    case GFX_KEYSF1:
		    case GFX_KEYSF2:
		    case GFX_KEYSF3:
		    case GFX_KEYSF4:
		    case GFX_KEYSF5:
		    case GFX_KEYSF6:
		    case GFX_KEYSF7:
		    case GFX_KEYSF8:
		    case GFX_KEYSF9:
		    case GFX_KEYSF10:
		    case GFX_KEYSF11:
		    case GFX_KEYSF12:
		    {
			int		whichfunc = key - GFX_KEYSF1;

			// Yell something!
			char buf[100];
			BUF		prompt;
			prompt.sprintf("Bind what to F%d? ", whichfunc+1);
			glbTalkBox->getString(prompt.buffer(), buf, 100);

			fkeys[whichfunc].strcpy(buf);

			prompt.sprintf("\"%s\" is now bound to F%d.",
					buf,
					whichfunc+1);

			glbTalkBox->appendText(prompt);
			didmove = false;
			break;
		    }
		    case '?':
		    {
			didmove = false;
			readBook(text_lookup("game::keys"));

			break;
		    }

		    case ' ':
		    {
			// Yell something!
			char buf[100];
			glbTalkBox->getString("Yell what? ", buf, 100);
			didmove = glbAvatar->actionYell(buf);

			glbTalkBox->clear();
			break;
		    }
		    case 'Q':
		    {
			// Verify they want to quit!
			char buf[100];
			glbTalkBox->getString("Are you sure you wish to awake? ", buf, 100);
			if (buf[0] == 'y' || buf[0] == 'Y')
			    done = true;
			glbTalkBox->clear();
			break;
		    }
		}
	    }
	}
	if (didmove)
	{
	    world->heartbeat();
	}
    }

    delete map;

    gfx_shutdown();
}
