#pragma once

#include "glbdef.h"
#include "pos.h"
#include "ptrlist.h"
#include "buf.h"
#include "intrinsic.h"

class MOB
{
public:
    MOB(MOB_NAMES defn);
    ~MOB();

    const MOB_DEF &defn() const
    {
	return glb_mobdefs[myDefinition];
    }
    MOB_NAMES		getDefinition() const { return myDefinition; }

    POS				pos() const { return myPos; }
    void			moveTo(POS newpos);

    void			die(const char *death);
    bool			isAvatar() const
    { return myDefinition == MOB_AVATAR; }
    bool			doAI();
    bool			heartbeat(int phase);
    bool			aiWander();
    bool			canMoveTo(POS pos) const;
    bool			canMoveIn(DIR dir) const { return canMoveTo(pos().delta(dir)); }

    bool			actionWalk(DIR dir);
    bool			actionAttack(DIR dir);
    bool			actionBump(DIR dir);
    bool			actionYell(const char *msg);
    bool			actionEmote(const char *msg);
    bool			actionCast(SPELL_NAMES spell);

    INTRINSIC_LIST		&intrinsic() { return myIntrinsics; }
    const INTRINSIC_LIST	&intrinsic() const { return myIntrinsics; }

    const char			*getMessage() const { return myMessage.buffer(); }
    const char			*getEmote() const { return myEmote.buffer(); }

    void			 clearMessages() { myMessage.clear(); myEmote.clear(); }

    DIR				getFacing() const { return myFacing; }
    void			setFacing(DIR facing) { myFacing = facing; }

protected:
    MOB_NAMES			myDefinition;	
    POS				myPos;
    MAP				*myDiskSourceMap;
    INTRINSIC_LIST		myIntrinsics;

    BUF				myMessage;
    BUF				myEmote;
    DIR				myFacing;
    DIR				myLastWalkDir;
    int				myLastPhase;
};

typedef PTRLIST<MOB *> MOBLIST;
